/* wydawca - automatic release submission daemon
   Copyright (C) 2007, 2008 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "sql.h"

/* Return the length of the string without trailing whitespace */
size_t
trim_length (const char *str)
{
  size_t len;

  for (len = strlen (str); len > 0 && isspace (str[len-1]); len--)
    ;
  return len;
}

/* Chop off any trailing whitespace. Return the length of the resulting string.
 */
size_t
trim (char *str)
{
  size_t len = trim_length (str);
  str[len] = 0;
  return len;
}

static const char *
expansion_value (struct kw_expansion *exp, void *data)
{
  if (exp->value)
    return exp->value;
  if (exp->expand)
    return exp->expand (exp, data);
  exp->value = "INTERNAL ERROR: NONEXPANDABLE DATA";
  exp->static_p = 1;
  return exp->value;
}

static const char *
find_expansion_char (int c, struct kw_expansion *exp, size_t nexp,
		     void *data)
{
  for (; nexp; nexp--, exp++)
    if (exp->kw[1] == 0 && exp->kw[0] == c)
      return expansion_value (exp, data);
  return NULL;
}

static const char *
find_expansion_word (char *kw, size_t len,
		     struct kw_expansion *exp, size_t nexp,
		     void *data)
{
  for (; nexp; nexp--, exp++)
    if (strlen (exp->kw) == len && memcmp (exp->kw, kw, len) == 0)
      return expansion_value (exp, data);
  return NULL;
}

/* Expand PARAM by replacing %u with the quoted value of USER, %p with that
   of PROJECT and %% with a single %. Return the malloc'ed result.
   FIXME: Should we quote PARAM itself?
*/
char *
expand_param (const char *param, struct kw_expansion *exp, size_t nexp,
	      void *data)
{
  char *p, *q, *res;
  const char *s;
  int len;
  
  if (!param)
    return NULL;
	
  /* Compute resulting param length */
  for (len = 0, p = (char *) param; *p; )
    {
      if (*p == '%')
	{
	  if (p[1] == '%')
	    {
	      len++;
	      p += 2;
	    }
	  else if (p[1] == '{')
	    {
	      char *e = strchr (p + 2, '}');
	      if (e && (s = find_expansion_word (p + 2, e - p - 2,
						 exp, nexp, data)))
		{
		  len += strlen (s);
		  p = e + 1;
		}
	      else 
		len++, p++;
	    }
	  else if ((s = find_expansion_char (p[1], exp, nexp, data)) != NULL)
	    {
	      len += strlen (s);
	      p += 2;
	    }
	  else 
	    {
	      /* FIXME: warning? */
	      len++;
	      p++;
	    }
	}
      else
	{
	  len++;
	  p++;
	}
    }
  
  res = xmalloc (len + 1);

  for (p = (char *) param, q = res; *p; )
    {
      if (*p == '%')
	{
	  switch (*++p)
	    {
	    case '%':
	      *q++ = *p++;
	      break;
	      
	    case '{':
	      {
		char *e = strchr (p + 1, '}');
		if (e
		    && (s = find_expansion_word (p + 1, e - p - 1,
						 exp, nexp, data)))
		  {
		    strcpy (q, s);
		    q += strlen (s);
		    p = e + 1;
		  }
		else 
		  len++, p++;
	      }
	      break;
	      
	    default:
	      if ((s = find_expansion_char (*p, exp, nexp, data)) != NULL)
		{
		  strcpy (q, s);
		  q += strlen (s);
		  p++;
		}
	      else
		*q++ = *p++;
	    }
	}
      else
	*q++ = *p++;
    }
  *q = 0;
  
  return res;
}

/* Quote non-printable characters in INPUT. Point *OUTPUT to the malloc'ed
   quoted string. Return its length. */
static size_t
quote_string (struct access_method *method, const char *input, char **poutput)
{
  size_t size, len;
  int quote;
  char *output;

  if (!input)
    {
      *poutput = xmalloc (1);
      (*poutput)[0] = 0;
      return 1;
    }
  
  switch (method->type)
    {
    case method_sql:
      len = strlen (input);
      size = 2 * len + 1;
      output = xmalloc (size);
      mysql_real_escape_string (&method->v.sqlconn->mysql, output, input, len);
      size = strlen (output);
      break;

    default:
      size = argcv_quoted_length (input, &quote);
      output = xmalloc (size);
      argcv_quote_copy (output, input);
      break;
    }

  *poutput = output;
  return size;
}

void
escape_kwexp (struct access_method *method, struct kw_expansion *exp,
	      size_t nexp)
{
  size_t i;

  for (i = 0; i < nexp; i++)
    {
      if (exp[i].value)
	{
	  quote_string (method, exp[i].value, &exp[i].value);
	  exp[i].static_p = 0;
	}
    }
}

void
free_kwexp (struct kw_expansion *exp, size_t nexp)
{
  size_t i;

  for (i = 0; i < nexp; i++)
    {
      if (!exp[i].static_p)
	free (exp[i].value);
      exp[i].value = NULL;
    }
}



int
fill_project_name (struct file_triplet *trp)
{
  char *project;
  const char *directory;
  char *p;

  if (trp->project)
    return 1;
  if (directive_get_value (trp, "directory", &directory))
    {
      logmsg (LOG_ERR, "%s: missing `directory' directive",
	      trp->file[file_directive].name);
      return 1;
    }
  p = strchr (directory, '/');
  if (p)
    {
      size_t len = p - directory;
      if (len == 0)
	{
	  logmsg (LOG_ERR, "%s: empty `directory' directive",
		  trp->file[file_directive].name);
	  return 1;
	}
      project = xmalloc (len + 1);
      memcpy (project, directory, len);
      project[len] = 0;
    }
  else
    project = xstrdup (directory);
  trp->project = project;
  return 0;
}

void
make_default_kwexp (struct kw_expansion kwexp[4], const char *user,
		    const char *project)
{
  memset (kwexp, 0, 4*sizeof(struct kw_expansion));
  kwexp[0].kw = "u";
  kwexp[0].value = (char*) user;
  kwexp[0].static_p = 1;
  kwexp[1].kw = "user";
  kwexp[1].value = (char*) user;
  kwexp[1].static_p = 1;
  kwexp[2].kw = "p";
  kwexp[2].value = (char*) project;
  kwexp[2].static_p = 1;
  kwexp[3].kw = "project";
  kwexp[3].value = (char*) project;
  kwexp[3].static_p = 1;
}  

/* Verify if USER has upload rights on the directory (project) requested
   by TRP */
int
check_access_rights (struct file_triplet *trp, struct directory_pair *dpair,
		     const char *user)
{
  struct access_method *method = dpair->verify_method;
  int rc;
  char *command;
  const char *result;
  struct kw_expansion kwexp[4];

  if (fill_project_name (trp))
    return 1;

  if (debug_level)
    logmsg (LOG_DEBUG, "verifying access rights for user %s to project %s",
	    user, trp->project);

  make_default_kwexp (kwexp, user, trp->project);
  
  escape_kwexp (method, kwexp, NITEMS (kwexp));
  command = expand_param (method->param[1], kwexp, NITEMS (kwexp), NULL);
  free_kwexp (kwexp, NITEMS (kwexp));
  
  rc = method_run (method, command);
  free (command);
  if (rc)
    {
      logmsg (LOG_ERR, "Cannot verify access rights for %s on %s",
	      user, trp->project);
      return 1;
    }

  result = method_result (method, 0, 0);
  if (!result || strcmp (result, user))
    {
      UPDATE_STATS (STAT_ACCESS_VIOLATIONS);
      logmsg (LOG_ERR, "%s: %s has no rights on %s",
	      trp->name, user, trp->project);
      notify (trp, ev_bad_ownership);
      return 1;
    }
	      
  return 0;
}

int
verify_directive_file (struct file_triplet *trp, struct directory_pair *dpair)
{
  struct passwd *pw;
  char *command;
  struct access_method *method = dpair->gpg_key_method;
  const char *pubkey;
  int rc;
  struct kw_expansion kwexp[4];
  
  if (!trp->file[file_directive].name)
    return 1;
  
  pw = getpwuid (TRIPLET_UID (trp));
  if (!pw)
    {
      logmsg (LOG_ERR, "%s: getpwuid failed: %s",
	      trp->name, strerror (errno));
      return 1;
    }
  trp->gid = pw->pw_gid;
  trp->user = xstrdup (pw->pw_name);

  make_default_kwexp (kwexp, trp->user, NULL);
  escape_kwexp (method, kwexp, NITEMS (kwexp));
  command = expand_param (method->param[1], kwexp, NITEMS (kwexp), NULL);
  free_kwexp (kwexp, NITEMS (kwexp));
  
  rc = method_run (method, command); 
  free (command);
  if (rc)
    {
      logmsg (LOG_ERR, "Cannot get PGP key for %s", trp->user);
      return 1;
    }
  
  pubkey = method_result (method, 0, 0);
  if (!pubkey || pubkey[0] == 0)
    {
      logmsg (LOG_ERR, "No GPG key for %s", trp->user);
      return 1;
    }

  if (debug_level > 3)
    logmsg (LOG_DEBUG, "Public key: %s", pubkey);

  if (verify_directive_signature (trp, dpair, pubkey))
    {
      /*FIXME: Update stats */
      logmsg (LOG_ERR, "invalid signature for %s", trp->name);
      return 1;
    }
  else if (debug_level)
    logmsg (LOG_DEBUG, "%s: directive file signature OK", trp->name);

  if (debug_level > 1)
    {
      int i;
      for (i = 0; trp->directive[i]; i++)
	logmsg (LOG_DEBUG, "directive[%d] = %s", i, trp->directive[i]);
    }

  if (verify_directive_format (trp))
    return 1;

  if (check_access_rights (trp, dpair, trp->user))
    return 1;

  return 0;
}

